#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <malloc.h>
#include <io.h>
#include <fcntl.h>
#include <math.h>
#include <mem.h>
#include <dos.h>
#include <time.h>
#include "misc.h"
#include "music.h"
#include "video.h"
#include "resource.h"
#ifdef __OS2__
#define INCL_DOS
#include <os2.h>
#else
#include "svgakit\ztimer.h"
#endif

int miscSoundCard=0, miscVideoMode=3;
int cosTable[cosSteps], sinTable[cosSteps];
int cosTable3D[cosSteps], sinTable3D[cosSteps];
int asinTable[asinSteps];

void error( char* msg ) {
#ifdef __OS2__
  WinAlarm( HWND_DESKTOP, WA_WARNING );
  OS2MsgBox(msg);
#else
  printf( "\nFatal: %s\n", msg );
#endif
  Shutdown();
  exit(-1);
}

void* operator new( size_t size )
{
  return cmalloc(size);
}

void operator delete( void* ptr )
{
  cfree(ptr);
}

static size_t allocated=0, max_alloc=0;

void* cmalloc( size_t size )
{
//  if( _heapchk() != _HEAPOK ) error( "Internal heap damage" );
  char* tmp = (char*)malloc(size+8);
  if(!tmp) error( "Not enough memory" );
  *(uint*)tmp = 0x12345678;
  *(uint*)(tmp+4)=size;
  allocated+=size; if( allocated>max_alloc) max_alloc=allocated;
//  printf( "cmalloc(%d), allocated=%d\n", size, allocated );
  return tmp+8;
}

void cfree( void* ptr )
{
  if(ptr) {
    if( *(((uint*)ptr)-2) != 0x12345678 ) error( "Heap corrupted" );
    uint size = *(((uint*)ptr)-1);
    free((char*)ptr-8);
    allocated -= size;
//    printf( "cfree(%d), allocated=%d\n", size, allocated );
  }
}

#ifdef __OS2__

//#define STACK_SIZE 4096

void miscAddTimerProc( miscTimerProc proc )
{
//  timer_procs.Insert((void*)_beginthread(proc,new char[STACK_SIZE],STACK_SIZE,0);
}

void miscRemoveTimerProc( miscTimerProc proc )
{
}

#else

TCollection timer_procs;
bool time_proc_busy=false;

void __interrupt __far (*old8) ();

void __interrupt __far timer_proc() {
  time_proc_busy = true;
  for( int i=0; i<timer_procs.Count; i++ ) {
    ((miscTimerProc)(timer_procs.At(i)))();
  }
  time_proc_busy = false;
  _chain_intr(old8);
}

void miscAddTimerProc( miscTimerProc proc )
{
  while(time_proc_busy);
  timer_procs.Insert(proc);
}

void miscRemoveTimerProc( miscTimerProc proc )
{
  while(time_proc_busy);
  timer_procs.Delete(proc);
}

#endif

void Startup( bool saveconfig )
{

  _fmode = O_BINARY;
  time_t foo;
  srand(time(&foo));

  for( int i=0; i<cosSteps; i++ ) {
    cosTable[i] = cos( 2*M_PI*i/cosSteps ) * (1<<cosScale);
    sinTable[i] = sin( 2*M_PI*i/cosSteps ) * (1<<cosScale);
    cosTable3D[i] = cos( 2*M_PI*i/cosSteps ) * (1<<cosScale3D);
    sinTable3D[i] = sin( 2*M_PI*i/cosSteps ) * (1<<cosScale3D);
  }
  for( i=0; i<asinSteps; i++ ) {
    asinTable[i] = asin((double)(i-asinSteps/2)/asinSteps)/M_PI*cosSteps*2;
  }

  musInitMusic();
  vidInitVideo();

#ifdef __OS2__
#else
  ZTimerInit();
  old8 = _dos_getvect( 8 );
  _dos_setvect( 8, timer_proc );
#endif

  if( saveconfig || !miscLoadConfig( PROJECT_NAME ".cfg" ) ) {
#ifdef __OS2__
#else
    cputs( PROJECT_NAME". (c)96 Queue Members Group Art Division\n\r\n\r" );
#endif
    miscSoundCard = musChooseCard();
    miscVideoMode = vidChooseVideoMode();
  }

  if( saveconfig ) {
    miscSaveConfig( PROJECT_NAME ".cfg" );
  }
  vidSetVideoMode(miscVideoMode);
  musInitCard(miscSoundCard);
}

void Shutdown()
{
  musStopMusic();
  musCloseMusic();
  vidCloseVideo();
#ifdef __OS2__
#else
  timer_procs.DeleteAll();
  _dos_setvect( 8, old8 );
  LZTimerOn();
  LZTimerOff();
  while( kbhit() ) getch();
  if( *(uchar*)0x417 & 3 ) {
    printf( "Memory allocated: now=%d, max=%d\n", allocated, max_alloc );
  }
#endif
}

void miscSaveConfig( char* filename ) {
  int h = creat( filename, 0 );
  if( h == -1 ) error( "Cannot write configuration file" );
  write( h, &miscSoundCard, sizeof(int) );
  write( h, &miscVideoMode, sizeof(int) );
  write( h, &vidHiResMode, sizeof(int) );
  close(h);
}

bool miscLoadConfig( char* filename )
{
  int h = open( filename, O_RDONLY );
  if( h == -1 )
    return false;
  else {
    read( h, &miscSoundCard, sizeof(int) );
    read( h, &miscVideoMode, sizeof(int) );
    read( h, &vidHiResMode, sizeof(int) );
    close(h);
    return true;
  }
}

#define ALLOC_STEP 10

static int (*_compare)(const void*, const void*);

int TCollection_Compare( const void* _1, const void* _2 )
{                           
  return _compare( *(void**)_1, *(void**)_2 );
}

void TCollection::Sort( int (*compare)(const void*, const void*) )
{
  _compare = compare;
  qsort( Items, Count, sizeof(void*), TCollection_Compare );
}

void TCollection::Insert( void* item )
{
  if( ++Count > allocated ) {
    allocated += ALLOC_STEP;
    void** tmp = Items;
    Items = (void**) cmalloc( sizeof(void*)*allocated );
    memcpy( Items, tmp, sizeof(void*)*(allocated-ALLOC_STEP) );
    cfree(tmp);
  }
  Items[Count-1] = item;
}

void TCollection::Delete(int index)
{
//  memcpy( &Items[index], &Items[index+1], sizeof(void*)*(allocated-index-1) );
//  Count--;
  if( index>=0 ) Items[index]=0;
}

void TCollection::PackDelete(int index)
{
  if( index>=0 ) {
    memcpy( &Items[index], &Items[index+1], sizeof(void*)*(allocated-index-1) );
    Count--;
  }
}

int TCollection::IndexOf( void* item )
{
  for( int i=0; i<Count; i++ ) {
    if( item == Items[i] ) return i;
  }
  return -1;
}

void TCollection::Pack()
{
  int i;
  while( (i=IndexOf(0)) != -1 ) PackDelete(i);
}

void TCollection::FreeAll()
{
  for( int i=0; i<Count; i++ ) {
    delete Items[i];
  }
  DeleteAll();
}

TCollection::~TCollection() {
  FreeAll();
  cfree(Items);
}

char* getToken(FILE* fp, const char* delimiters )
{
  static char buf[1024];
  char* ptr = buf;
  do {
    *ptr = fgetc(fp);
  } while ( strchr( delimiters, *ptr ) );
  do {
    *(++ptr) = fgetc(fp);
  } while ( !strchr( delimiters, *ptr ) && (ptr-buf)<1023 );
  *ptr = 0;
//  printf( "%s\n", buf );
  return buf;
}

#ifdef __OS2__

uint timerOffset=0;

uint OS2GetTimer()
{
  DATETIME dt;
  DosGetDateTime(&dt);
  return dt.hours*60*60*1000+dt.minutes*60*1000+dt.seconds*1000+dt.hundredths*10;
}

#endif

void miscStartTimer()
{
#ifdef __OS2__
  timerOffset = OS2GetTimer();
#else
  LZTimerOn();
#endif
}

uint miscTimer()
{
#ifdef __OS2__
  return OS2GetTimer()-timerOffset;
#else
  return LZTimerLap()*LZTIMER_RES*miscTimerRes;
#endif
}

void _memsetw(  void* dst, ushort value, size_t length );
#pragma aux _memsetw = \
"push ax"\
"shl eax,10h"\
"pop ax"\
"shr ecx,1"\
"rep stosd"\
parm [edi] [ax] [ecx];

void memsetw( void* dst, ushort value, size_t length )
{
  _memsetw( dst, value, length );
}
